#include <l4.h>

#include "misc.h"
#include "task.h"
#include "transaction.h"


#define CASE(request, ntyped, nuntyped)				\
case request:							\
  if (tag.typed != (ntyped) || tag.untyped != (nuntyped))	\
    { err = ERR_BAD_ARGS; goto fail; }

/* Main function for each server thread. */
static void
serve(void)
{
  for (;;)
    {
      l4_word_t err;
      l4_msg_tag_t tag;
      l4_thread_id_t caller;
      l4_word_t caller_task;
      
      tag = l4_wait(&caller);
      if (l4_ipc_failed(tag))
	continue;

      /* Checks that the caller is in the right subsystem, and has a
	 sane task id */
      if (!(caller_task = task_by_thread(caller)))
	/* Should we really send a reply, or is it better to just drop
	   the request on the floor ? */
	{ err = ERR_GO_AWAY; goto fail; }

      switch(l4_label(tag))
	{
	  CASE (TASK_CREATE, 2, 0)
	    {
	      l4_thread_id_t pager;
	      l4_word_t thread_max;
	      l4_word_t new_task; /* Control handle for new task */
	    
	      l4_store_mr(1, &pager.raw);
	      l4_store_mr(2, &thread_max);
	      err = task_create(caller_task, thread_max, &new_task);

	      if (err)
		goto fail;

	      l4_set_msg_tag(make_l4_tag(0, 1, 0));
	      l4_load_mr(1, new_task);

	      break;
	    }
	  CASE (THREAD_CREATE, 2, 0)
	    {
	      l4_word_t handle;
	      l4_thread_id_t pager;
	      l4_word_t id;
	      l4_thread_id_t new_thread;
	      
	      l4_store_mr(1, &handle);
	      l4_store_mr(2, &pager.raw);

	      id = task_by_control_handle(caller_task, handle);
	      if (!id)
		{ err = ERR_PERM; goto fail; }

	      err = task_thread_create(id, pager, &new_thread);
	      if (err)
		goto fail;

	      l4_set_msg_tag(make_l4_tag(0, 1, 0));
	      l4_load_mr(1, new_thread.raw);

	      break;
	    }
	  CASE (THREAD_DESTROY, 2, 0)
	    {
	      l4_word_t handle;
	      l4_thread_id_t victim;

	      l4_word_t id;

	      l4_store_mr(1, &handle);
	      l4_store_mr(2, &victim.raw);

	      id = task_by_control_handle(caller_task, handle);
	      if (!id)
		{ err = ERR_PERM; goto fail; }

	      err = task_thread_destroy(id, victim);
	      if (err)
		goto fail;
	      l4_set_msg_tag(make_l4_tag(0, 0, 0));

	      break;
	    }
	      
	  CASE (TRANSACTION_SEND_CREATE, 3, 0)
	    {
	      /* Initiates sending HANDLE from CALLER_TASK to
		 RECEIVER_TASK. SUBJECT is the task that HANDLE refers
		 to. */
	      l4_word_t receiver;
	      l4_word_t handle;
	      l4_word_t control;
	      
	      l4_word_t subject;
	      l4_word_t receiver_task;
	      l4_word_t new_transaction;
	      
	      l4_store_mr(1, &receiver);
	      l4_store_mr(2, &handle);
	      l4_store_mr(3, &control);
	      
	      receiver_task = task_by_handle(caller_task, receiver);
	      if (!receiver_task)
		{ err = ERR_NO_TASK; goto fail; }

	      subject = task_by_handle(caller_task, handle);
	      if (!subject)
		{ err = ERR_NO_TASK; goto fail; }

	      /* Modify the handle that the receiver gets */
	      if (!handle)
		handle = (control
			  ? task_control_handle(subject)
			  : task_reference_handle(subject));
	      else if (!control)
		handle &= ~TASK_CONTROL_MASK;
	      
	      err = transaction_create_send_record(caller_task,
						   receiver_task,
						   subject,
						   handle,
						   &new_transaction);
	      if (err)
		goto fail;
	      
	      l4_set_msg_tag(make_l4_tag(0, 1, 0));
	      l4_load_mr(1, new_transaction);

	      break;	      
	    }
	  CASE (TRANSACTION_RECEIVE, 2, 0)
	    {
	      /* The CALLER tries to get a handle from SENDER, using
	         the transaction id prviously created by the sender */
	      l4_word_t sender;
	      l4_word_t transaction;

	      l4_word_t sender_task;
	      l4_word_t handle;
	      
	      l4_store_mr(1, &sender);
	      l4_store_mr(2, &transaction);

	      sender_task = task_by_handle(caller_task, sender);
	      if (!sender_task)
		{ err = ERR_NO_TASK; goto fail; }
	      
	      err = transaction_receive(transaction,
					sender_task,
					caller_task,
					&handle);

	      if (err)
		goto fail;
	      
	      l4_set_msg_tag(make_l4_tag(0, 1, 0));
	      l4_load_mr(1, handle);

	      break;
	    }
	  CASE (TRANSACTION_DESTROY, 1, 0)
	    {
	      l4_word_t transaction;
	      
	      l4_store_mr(1, &transaction);

	      err = transaction_destroy(transaction, caller_task);
	      if (err)
		goto fail;

	      l4_set_msg_tag(make_l4_tag(0, 0, 0));
	    }
	default:
	  err = ERR_UNKNOWN_REQUEST;
	fail:
	  l4_set_msg_tag(make_l4_tag(err, 0, 0));
	  break;
	}
      
      /* Ignores return value */
      (void) l4_reply(caller);
    }
}

int
main(void)
{
  task_init();
  transaction_init();
  
  /* Only one thread, for now */
  serve();

  return 1;
}
