#include "bitmap.h"
#include "misc.h"
#include "task.h"
#include "transaction.h"


/* Should be declared in <string.h> and <assert.h> */
#define memset __builtin_memset
#define assert(x)

static l4_fpage_t kip_fpage;

/* Our subsystem */
static l4_word_t subsystem;

/* The implementation wastes some memory, with mostly empty bitmaps.
 * But it's an easy way to keep track of everything, using only static
 * allocation. */
struct task_object
{
  enum { TASK_DEAD = 0,
	 TASK_EMPTY,
	 TASK_ALIVE,
	 TASK_ZOMBIE } state;

  /* An inactive thread (tid zero) used to identify the space. */
  l4_thread_id_t space;
  
  /* For now, only a single accounting id. Should be a list of
   * handles of the form <owner, id>. */
  l4_word_t accounting_id;

  /* Number of active threads */
  l4_word_t nthreads;
  /* Number of threads that we have utc blocks for */
  l4_word_t thread_max;
  l4_word_t thread_map[THREAD_MAX / L4_BITS_PER_WORD];

   /* Includes both ordinary references and
      control references */
  l4_word_t nreferrers;

  l4_word_t referrer_map[TASK_MAX / L4_BITS_PER_WORD];
  l4_word_t controller_map[TASK_MAX / L4_BITS_PER_WORD];

  /* Number of pending transactions for which this task is a subject.
     A task zombie that is subject to a handle transaction is
     preserved even if it has no other references. */
  l4_word_t ntransactions;
  
  /* Set if sending a death notification to this task fails */
  int out_of_sync;
  l4_thread_id_t notify;
};

/* A task id is an index into this array. Entries with index divisable
   by 64 are not used (as the task id:s would collide with local
   thread id:s). */
static struct task_object
task[TASK_MAX];

void
task_init(void)
{
  /* Redundant if linker/loader does its job */
  memset(task, 0, sizeof(task));

  subsystem = l4_thread_no(l4_myself()) >> SUBSYSTEM_SHIFT;
  
  /* The kip is always one page, right? */
  kip_fpage = l4_fpage(KIP_LOCATION,
		       l4_page_size_mask());
}

#define TASK_ID_ALLOWED(id) (((id) % 64) != 0)

/* Finds a free task id. Doesn't mark it as used, the caller has to do
   that. */
l4_word_t
task_alloc_id(void)
{
  /* Inefficient linear search */
  l4_word_t id;
  for (id = 2; id < TASK_MAX; id++)
    if (TASK_ID_ALLOWED(id) && task[id].state == TASK_DEAD)
      return id;
  
  return 0;
}

static void
task_maybe_dealloc(l4_word_t id)
{
  if (task[id].state == TASK_ZOMBIE
      && !task[id].nreferrers
      && !task[id].ntransactions)
    task[id].state = TASK_DEAD;
}

/* Returns non-zero if task A has a reference to task B */
l4_word_t
task_reference_p(l4_word_t a, l4_word_t b)
{
  return bit_p(a, task[b].referrer_map);
}

void
task_reference_add(l4_word_t a, l4_word_t b)
{
  if (!task_reference_p(a,b))
    {
      bit_set(a, task[b].referrer_map);
      task[b].nreferrers++;
    }
}

void
task_reference_clear(l4_word_t a, l4_word_t b)
{
  if (task_reference_p(a,b))
    {
      assert(task[b].referrers);

      bit_clear(a, task[b].referrer_map);

      if (!--task[b].nreferrers)
	task_maybe_dealloc(b);
    }
}

/* Returns non-zero if task A can control task B */
l4_word_t
task_control_p(l4_word_t a, l4_word_t b)
{
  return bit_p(a, task[b].controller_map);
}

void
task_control_add(l4_word_t a, l4_word_t b)
{
  if (!task_control_p(a,b))
    {
      bit_set(a, task[b].controller_map);
      task[b].nreferrers++;
    }
}

/* Returns zero if this was the last reference to B. */
void
task_control_clear(l4_word_t a, l4_word_t b)
{
  if (task_control_p(a,b))
    {
      assert(task[b].referrers);

      bit_clear(a, task[b].controller_map);
      
      if (!--task[b].nreferrers)
	task_maybe_dealloc(b);
    }
}

void
task_transaction_incref(l4_word_t id)
{
  assert(task[id].state != TASK_DEAD);
  task[id].ntransactions++;
}

void
task_transaction_decref(l4_word_t id)
{
  assert(task[id].state != TASK_DEAD);
  assert(task[id].ntransactions);
  task[id].ntransactions--;

  task_maybe_dealloc(id);
}


l4_word_t
task_reference_handle(l4_word_t id)
{
  return id;
}

l4_word_t
task_control_handle(l4_word_t id)
{
  return id | TASK_CONTROL_MASK;
}


/* Functions returning thread id:s */
l4_word_t
task_by_thread(l4_thread_id_t thread)
{
  l4_word_t id;
  
  if (!l4_is_global_id(thread)
      || (l4_thread_no(thread) >> SUBSYSTEM_SHIFT) != subsystem)
    return 0;

  id = l4_version(thread);
  /* id 1 is reserved for the task server itself */
  if (1 < id && id < TASK_MAX)
    return id;

  return 0;
}

/* Checks that the handle has the right syntax, and refers to a
   non-dead task. Ignores the control bit. */
static l4_word_t
task_check_handle(l4_word_t handle)
{
  l4_word_t id;
  
  if (handle / (2*TASK_CONTROL_MASK))
    return 0;

  id = handle % TASK_CONTROL_MASK;
  if (handle < TASK_MAX
      /* Somewhat redundant, as the state should be TASK_DEAD */
      && TASK_ID_ALLOWED(id)
      && task[id].state != TASK_DEAD)
    return id;
  else
    return 0;
}

l4_word_t
task_by_handle(l4_word_t owner, l4_word_t handle)
{
  l4_word_t id;

  /* Special handle 0 is a control handle of the owner. */
  if (!handle)
    return owner;
  
  if ( (id = task_check_handle(handle))
       && (task_reference_p(owner, id)
	   || task_control_p(owner, id)))
    return id;
  else
    return 0;
}

l4_word_t
task_by_control_handle(l4_word_t owner, l4_word_t handle)
{
  l4_word_t id;

  /* Special handle 0 is a control handle of the owner. */
  if (!handle)
    return owner;

  if ( (handle & TASK_CONTROL_MASK)
       && (id = task_check_handle(handle))
       && task_control_p(owner, id))
    return id;
  else
    return 0;
}


/* Thread allocation */

static inline int
task_thread_used_p(l4_word_t id, l4_word_t tid)
{
  return bit_p(tid, task[id].thread_map);
}

static l4_word_t
task_thread_alloc(l4_word_t id, void **utc)
{
  /* Again, an unefficient linear search */
  unsigned tid;

  /* tid zero is reserved for the inactive thread used as the space
   * specifier */
  for (tid = 1; id < task[id].thread_max; id++)
    if (task_thread_used_p(id, tid))
      {
	/* A free thread number */
	*utc = (void *) (UTC_LOCATION + tid * l4_utcb_size());
	return tid;
      }
  return 0;
}

static void
task_thread_use(l4_word_t id, l4_word_t tid)
{
  assert(!task_thread_used_p(id, tid));
  bit_set(tid, task[id].thread_map);
  task[id].nthreads++;
}

static void
task_thread_unuse(l4_word_t id, l4_word_t tid)
{
  assert(task[id].nthreads);
  assert(task_thread_used_p(id, tid));
  bit_clear(tid, task[id].thread_map);
  task[id].nthreads--;
}

static inline l4_thread_id_t
make_thread_id(l4_word_t id, l4_word_t tid)
{
  return l4_global_id(tid + (subsystem << SUBSYSTEM_SHIFT),
		      id); 
}

/* Checks that THREAD really is a thread in the given task,
   and returns the tid. */
static inline l4_word_t
check_thread_id(l4_word_t id, l4_thread_id_t thread)
{
  l4_word_t tid;
  
  if (id != task_by_thread(thread))
    return 0;
  
  tid = l4_thread_no(thread) % (1UL << SUBSYSTEM_SHIFT);

  if (task_thread_used_p(id, tid))
    return tid;
  else
    return 0;
}


/* Notifications */

static void
task_zombify(l4_word_t id)
{
  l4_word_t o;
  assert(task[id].state == TASK_ALIVE);
  assert(!task[id].nthreads);

  /* Delete the address space. Ignores any error. */
  (void) l4_thread_control(task[id].space, l4_nilthread, l4_nilthread,
			   l4_nilthread, (void *) -1);
  
  /* First clean up any transaction references the dead task had. The
     call may deallocate some zombie tasks, so we must keep our state
     as alive during this call. */
  transaction_note_death(id);

  task[id].state = TASK_ZOMBIE;
  
  /* References we have to other tasks should be deleted */
  for (o = 2; o < TASK_MAX; o++)
    {
      if (task[o].state != TASK_DEAD)
	{
	  if (!l4_is_nil_thread(task[o].notify)
	      && (task_reference_p(id, o)
		  || task_control_p(id, o)))
	    {
	      /* We send a single timeout, no matter if the task had a
		 reference or a control handle or both. We don't do
		 any retransmissions, but we record failures. A task
		 could find out that it is out of sync and request a
		 list of all handles it has to zombie tasks. */
		 
	      l4_set_msg_tag(make_l4_tag(TASK_DEATH_NOTIFICATION, 1, 0));
	      l4_load_mr(1, id);

	      if (l4_ipc_failed(l4_reply(task[o].notify)))
		task[o].out_of_sync = 1;
	    }
	  task_reference_clear(id, o);
	  task_control_clear(id, o);
	}
    }
  
  task_maybe_dealloc(id);
}

/* Operations */
l4_word_t
task_create(l4_word_t creator, l4_word_t thread_max, l4_word_t *handle)
{
  l4_word_t id = task_alloc_id();
  l4_thread_id_t space;
  l4_word_t old;
  
  if (!id)
    return ERR_NOMEM;

  if (!thread_max || thread_max > THREAD_MAX)
    thread_max = THREAD_MAX;
  else
    thread_max = thread_max;
  
  /* Create an address space */
  space = make_thread_id(id, 0);
  
  if (!l4_thread_control(/* thread, space, scheduler */
			 space, space, space,
			 /* pager, utc */
			 l4_nilthread, (void *) -1))
    return ERR_NOMEM;

  if (!l4_space_control(space, 0,
			kip_fpage,
			l4_fpage(UTC_LOCATION,
				 /* FIXME: Alignment. */
				 thread_max * l4_utcb_size()),
			l4_anythread, &old))
    {
      /* Ignores any error */
      l4_thread_control(/* thread, space, scheduler */
			space, l4_nilthread, space,
			/* pager, utc */
			l4_nilthread, (void *) -1);
      return ERR_NOMEM;
    }

  task[id].space = space;
  task[id].thread_max = thread_max;
  task[id].state = TASK_EMPTY;
  task_control_add(creator, id);

  /* How should the child get to know this handle? */
  task_reference_add(id, creator);

  *handle = task_control_handle(id);
  return 0;
}

l4_word_t
task_thread_create(l4_word_t id, l4_thread_id_t pager,
		   l4_thread_id_t *new_thread)
{
  l4_word_t tid;
  l4_thread_id_t thread;
  void *utc;

  tid = task_thread_alloc(id, &utc);
  if (!tid)
    return ERR_NOMEM;

  thread = make_thread_id(id, tid);
  
  if (task[id].state == TASK_ZOMBIE)
    return ERR_STATE;

  if (!l4_thread_control(/* thread, space, scheduler */
			 thread, task[id].space, thread,
			 /* pager, utc */
			 pager, utc))
    return ERR_NOMEM;

  task_thread_use(id, tid);
  task[id].state = TASK_ALIVE;
  
  *new_thread = thread;
  return 0;
}

l4_word_t
task_thread_destroy(l4_word_t id, l4_thread_id_t victim)
{
  l4_word_t tid = check_thread_id(id, victim);

  if (!tid)
    return ERR_NO_THREAD;
  
  if (!l4_thread_control(victim, l4_nilthread, l4_nilthread,
			 l4_nilthread, (void *) -1))
    return ERR_FAIL;

  task_thread_unuse(id, tid);
  if (!task[id].nthreads)
    task_zombify(id);
  
  return 0;
}
