#include "task.h"

/* Should be declared in <string.h> and <assert.h> */
#define memset __builtin_memset
#define assert(x)

#define TRANSACTION_MAX 1000

/* A transaction record, for sending a task handle */
struct transaction_record
{
  enum transaction_state
    { TRANSACTION_FREE = 0,
      TRANSACTION_PENDING,
      TRANSACTION_ERROR,
      TRANSACTION_DONE } state;
  l4_word_t sender;
  l4_word_t receiver;
  l4_word_t subject;
  l4_word_t handle;
};

/* A transaction record imples an automatic reference to the subject
   of the transaction. That makes it possible to pass around
   references to zombies. A transaction record does not imply any
   reference to the sender or receiver; if the receiver dies, the
   record state changes to TRANSACTION_ERROR. If the sender dies, the
   record is deallocated immediately. */

/* A transaction id is an index into this array. Index zero is not
   used. */
static struct transaction_record
transaction[TRANSACTION_MAX];

void
transaction_init(void)
{
  /* Redundant if linker/loader does its job */
  memset(transaction, 0, sizeof(transaction));
}

l4_word_t
transaction_alloc(void)
{
  /* Inefficient linear search */
  l4_word_t id;
  for (id = 1; id < TRANSACTION_MAX; id++)
    if (transaction[id].state == TRANSACTION_FREE)
      return id;

  return 0;
}


/* Called when task id turns into a zombie */
void
transaction_note_death(l4_word_t task)
{
  l4_word_t id;
  for (id = 1; id < TRANSACTION_MAX; id++)
    {
      if (transaction[id].state == TRANSACTION_FREE)
	;
      
      /* If the sender is gone, the transaction is deallocated
	 immediately */
      else if (transaction[id].sender == task)
	{
	  if (transaction[id].state == TRANSACTION_PENDING)
	    task_transaction_decref(transaction[id].subject);
	  
	  transaction[id].state = TRANSACTION_FREE;
	  continue;
	}
      /* Receiver death matters only for pending transactions */
      else if (transaction[id].state == TRANSACTION_PENDING
	       && transaction[id].receiver == task)
	{
	  transaction[id].state = TRANSACTION_ERROR;
	  task_transaction_decref(transaction[id].subject);
	}
    }
}

/* Operations */

l4_word_t
transaction_create_send_record(l4_word_t sender,
			       l4_word_t receiver,
			       l4_word_t subject,
			       l4_word_t handle,
			       l4_word_t *new_transaction)
{
  l4_word_t id = transaction_alloc();
  if (!id)
    return ERR_NOMEM;

  task_transaction_incref(subject);

  transaction[id].state = TRANSACTION_PENDING;
  transaction[id].sender = sender;
  transaction[id].receiver = receiver;
  transaction[id].subject = subject;
  transaction[id].handle = handle;
  
  *new_transaction = id;
  return 0;
}

l4_word_t
transaction_receive(l4_word_t id,
		    l4_word_t sender,
		    l4_word_t receiver,
		    l4_word_t *handle)
{
  if (id >= TRANSACTION_MAX
      || transaction[id].state == TRANSACTION_FREE)
    return ERR_NO_TRANSACTION;

  /* It's not useful to distinguish between these errors,
     as any state is possible if the sender has destroyed the record. */
  if (id && id < TRANSACTION_MAX
      && transaction[id].state == TRANSACTION_PENDING
      && transaction[id].sender == sender
      && transaction[id].receiver == receiver)
    {
      l4_word_t subject = transaction[id].subject;
      l4_word_t err = 0;
            
      if (transaction[id].handle & TASK_CONTROL_MASK)
	task_control_add(receiver, subject);

      else
	task_reference_add(receiver, subject);
      
      transaction[id].state = TRANSACTION_DONE;
      *handle = transaction[id].handle;

      return err;
    }
  else
    return ERR_STATE;
}

l4_word_t
transaction_destroy(l4_word_t id, l4_word_t sender)
{
  enum transaction_state state;
  
  if (id >= TRANSACTION_MAX
      || transaction[id].state == TRANSACTION_FREE
      || transaction[id].sender != sender)
    return ERR_NO_TRANSACTION;

  state = transaction[id].state;
  transaction[id].state = TRANSACTION_FREE;

  switch (state)
    {
    case TRANSACTION_PENDING:
      task_transaction_decref(transaction[id].subject);
      return ERR_TRANSACTION_INTERRUPTED;
    case TRANSACTION_ERROR:
      return ERR_TRANSACTION_FAIL;
    case TRANSACTION_DONE:
      return 0;
    default:
      /* Internal error */
      /* abort(); */
      return ERR_FAIL;
    }
}
